from zedsuite.utils import RANDOM_NAME, CythonWrapper
import ctypes
from cython cimport sizeof
from zedsuite.genesess cimport _genESeSS
from libc.stdlib cimport malloc, free
from libc.string cimport strcpy


class GenESeSS(CythonWrapper):
    def __init__(
        self,
        helpmsg: bool = False,
        version: bool = False,
        configfile: str = None,
        datafile: str = None,
        data_len: ctypes.c_uint = None, # len in genESeSS.cc
        data_type: str = None,
        data_dir: str = None,
        partition: list = None,
        eps: ctypes.c_double = None,
        timer: int = 0,
        derivative: bool = None,
        force: bool = None,
        outfile: str = None,
        dotcfg: str = None,
        dot: str = None,
        sym_frq_display_depth_: ctypes.c_uint = None,
        featurization: bool = None,
        runl: ctypes.c_uint = None,
        modelruns: ctypes.c_uint = None,
        runfile: str = None,
        symderiv: str = None,
        allsym: str = None,
        verbose: int = 0
    ) -> None:

        self.argc = 1
        self.argv = ['genESeSS']
        self.program_info = False

        # Program information:
        if helpmsg:
            self.program_info = True
            self._add_flag('-h')
        if version:
            self.program_info = True
            self._add_flag('-V')

        # Usage:
        if configfile is not None:
            self._add_param('-c', configfile)
        if datafile is not None:
            self._add_param('-f', datafile)
        if data_len is not None:
            self._add_param('-x', str(data_len))
        if data_type is not None:
            self._add_param('-T', data_type)
        if data_dir is not None:
            self._add_param('-D', data_dir)
        if partition is not None:
            self._add_param('-P', ' '.join([str(n) for n in partition]))
        if eps is not None:
            self._add_param('-e', str(eps))
        if timer is not None:
            self._add_param('-t', str(timer))
        if derivative is not None:
            self._add_param('-u', str(int(derivative)))
        if force is not None:
            self._add_flag('-F')

        # Output options:
        self.outfile = None
        if outfile is not None:
            self.outfile = outfile
            self._add_param('-o', self.outfile)
        else:
            self.outfile = RANDOM_NAME(path=self.temp_dir)
            self._add_param('-o', self.outfile)
        if dotcfg is not None:
            self._add_param('-d', dotcfg)
        if dot is not None:
            self._add_param('-d', dot)
        if sym_frq_display_depth_ is not None:
            self._add_param('-W', str(sym_frq_display_depth_))
        if featurization is not None:
            self._add_param('-y', str(int(featurization)))
        if runl is not None:
            self._add_param('-r', str(runl))
        if modelruns is not None:
            self._add_param('-N', str(modelruns))
        if runfile is not None:
            self._add_param('-R', runfile)
        if symderiv is not None:
            self._add_param('-S', symderiv)
        if allsym is not None:
            self._add_param('-H', allsym)
        if verbose is not None:
            self._add_param('-v', str(verbose))

        # Result variables:
        self.inference_error = None
        self.epsilon_used = None
        self.synchronizing_string_found = None
        self.symbol_frequency = None
        self.probability_morph_matrix = None
        self.connectivity_matrix = None

    def run(self) -> None:
        # convert self.argv to a char** to be passed to main()
        c_argv = <char **> malloc(self.argc * sizeof(ctypes.c_char_p))
        for i, arg in enumerate(self.argv):
            c_argv[i] = <char *> malloc(len(arg) * sizeof(char))
            strcpy(c_argv[i], bytes(arg, 'utf-8'))

        # _genESeSS() defined in genESeSS.cc
        _genESeSS(self.argc, c_argv)

        for i in range(self.argc):
            free(c_argv[i])
        free(c_argv)

        # if passed help or version flags
        if self.program_info:
            return


        with open(self.outfile, 'r') as f:
            for line in f:
                split = line.split(':')

                if split[0] == '%ANN_ERR':
                    self.inference_error = float(split[1].strip())
                elif split[0] == '%MRG_EPS':
                    self.epsilon_used = float(split[1].strip())
                elif split[0] == '%SYN_STR':
                    try:
                        self.synchronizing_string_found = [int(val.strip()) for val in split[1].strip().split(' ')]
                    except:
                        self.synchronizing_string_found = None
                elif split[0] == '%SYM_FRQ':
                    self.symbol_frequency = [float(val.strip()) for val in split[1].strip().split(' ')]
                elif split[0] == '%PITILDE':
                    size = int(split[1].strip().split('(')[1].split(')')[0])
                    self.probability_morph_matrix = []
                    next(f) # skip #PITILDE line
                    for i in range(size):
                        self.probability_morph_matrix.append([float(val) for val in next(f).strip().split(' ')])
                elif split[0] == '%CONNX':
                    size = int(split[1].strip().split('(')[1].split(')')[0])
                    self.connectivity_matrix = []
                    next(f) # skip #CONNX line
                    for i in range(size):
                        self.connectivity_matrix.append([int(val) for val in next(f).strip().split(' ')])



# details here: https://stackoverflow.com/a/52714500/8728279
# bootstrap is the only module which
# can be loaded with default Python-machinery
# because the resulting extension is called `bootstrap`:
from . import bootstrap

# injecting our finders into sys.meta_path
# after that all other submodules can be loaded
bootstrap.bootstrap_cython_submodules()


import os
import ctypes
import pandas as pd
from zedsuite.utils import RANDOM_NAME, CythonWrapper
from cython import boundscheck, nonecheck, wraparound, cdivision
from cython cimport sizeof
from zedsuite.zutil cimport _lsmash
from libc.stdlib cimport malloc, free
from libc.string cimport strcpy


class Llk(CythonWrapper):
    def __init__(
        self,
        helpmsg: bool = False,
        version: bool = False,
        seq: pd.DataFrame = None,
        seqfile: str = None,
        data_dir: str = None,
        data_len: ctypes.c_uint = None,
        pfsafile: str = None
    ) -> None:

        super().__init__()

        self.argc = 1
        self.argv = ['llk']

        # Program information:
        if helpmsg:
            self._add_flag("-h")
        if version:
            self._add_flag("-V")

        # Usage:
        self.seqfile = None
        if seq is not None:
            self.seqfile = os.path.abspath(RANDOM_NAME(path=self.temp_dir))
            seq.to_csv(
                self.seqfile,
                sep=" ",
                line_terminator=(" " if type(seq) is pd.Series else "\n"),
                index=False,
                header=False
            )
            self._add_param("-s", self.seqfile)
        if seqfile is not None:
            self._add_param("-s", seqfile)
        if data_dir is not None:
            self._add_param("-D", data_dir)
        if data_len is not None:
            self._add_param("-x", str(data_len))
        if pfsafile is not None:
            self._add_param("-f", pfsafile)

        # Results
        self.llk_vec = None

    def run(self) -> [float]:
        # convert self.argv to a char** to be passed to main()
        argv = <char **> malloc(self.argc * sizeof(ctypes.c_char_p))
        for i, arg in enumerate(self.argv):
            argv[i] = <char *> malloc(len(arg) * sizeof(char))
            strcpy(argv[i], bytes(arg, 'utf-8'))

        # llk() defined in _llk.cc
        self.llk_vec = _llk(self.argc, argv)
        # cdef unsigned int argc = self.argc
        # cdef vector[double] llk_vec
        # openmp.omp_set_dynamic(1)
        # with nogil, parallel():
        #     llk_vec = _llk(argc, argv)
        # self.llk_vec = llk_vec

        for i in range(self.argc):
            free(argv[i])
        free(argv)

        return self.llk_vec


class Lsmash(CythonWrapper):
    def __init__(
        self,
        helpmsg: bool = False,
        version: bool = False,
        seq: pd.DataFrame = None,
        seqfile: str = None,
        data_dir: str = None,
        data_type: str = None,
        data_len: ctypes.c_uint = None,
        partition: list = None,
        derivative: bool = None,
        pfsafiles: list = None,
        timer: bool = None,
        sae: bool = None,
        repeat: ctypes.c_uint = None,
        outfile: str = None,
        random_mc: ctypes.c_uint = None,
        print_mc: bool = None,
    ) -> None:

        super().__init__()

        self.argc = 1
        self.argv = ['lsmash']

        # Program information:
        if helpmsg:
            self._add_flag("-h")
        if version:
            self._add_flag("-V")

        # Usage:
        self.seqfile = None
        if seq is not None:
            self.seqfile = os.path.abspath(RANDOM_NAME(path=self.temp_dir))
            seq.to_csv(self.seqfile, sep=" ", index=False, header=False)
            self._add_param("-f", self.seqfile)
        if seqfile is not None:
            self._add_param("-f", seqfile)
        if data_dir is not None:
            self._add_param("-D", data_dir)
        if data_type is not None:
            self._add_param("-T", data_type)
        if data_len is not None:
            self._add_param("-x", str(data_len))
        if partition is not None:
            self._add_param("-P", " ".join([str(n) for n in partition]))
        if derivative is not None:
            self._add_param("-u", str(derivative))
        if pfsafiles is not None:
            self._add_param("-F", " ".join([str(n) for n in pfsafiles]))
        if timer is not None:
            self._add_param("-t", str(timer))
        if sae is not None:
            self._add_param("-S", str(sae))
        if repeat is not None:
            self._add_param("-n", str(repeat))
        if outfile is not None:
            self._add_param("-o", outfile)
        if random_mc is not None:
            self._add_param("-R", str(random_mc))
        if print_mc is not None:
            self._add_param("-m", str(print_mc))

        # Results
        self.dist_matrix = None

    @boundscheck(False)
    @nonecheck(False)
    @wraparound(False)
    @cdivision(True)
    def run(self) -> pd.DataFrame:
        # convert self.argv to a char** to be passed to main()
        cdef char** argv = <char **> malloc(self.argc * sizeof(ctypes.c_char_p))
        for i, arg in enumerate(self.argv):
            argv[i] = <char *> malloc(len(arg) * sizeof(char))
            strcpy(argv[i], bytes(arg, 'utf-8'))

        # lsmash() defined in _lsmash.cc
        self.dist_matrix = _lsmash(self.argc, argv)

        # testing parallelization without GIL
        # cdef unsigned int argc = self.argc
        # cdef matrix_dbl dist_matrix
        # openmp.omp_set_dynamic(1)
        # with nogil, parallel():
        #     dist_matrix = _lsmash(argc, argv)
        # self.dist_matrix = dist_matrix

        for i in range(self.argc):
            free(argv[i])
        free(argv)

        return self.dist_matrix



import pandas as pd
import shutil
import uuid
import os
import atexit
import traceback
import glob
import inspect

class CythonWrapper:
    def __init__(self):
        self.temp_dir = "zed_temp"

    def _add_flag(self, flag: str) -> None:
        self.argv.append(flag)
        self.argc = len(self.argv)

    def _add_param(self, flag: str, value: str) -> None:
        self.argv.append(flag)
        self.argv.append(value)
        self.argc = len(self.argv)


def os_remove(filename):
    try:
        os.remove(filename)
    except OSError:
        pass


def run_once_per_process(f):
    def wrapper(*args, **kwargs):
        if os.getpgrp() != wrapper.has_run:
            wrapper.has_run = os.getpgrp()
            return f(*args, **kwargs)

    wrapper.has_run = 0
    return wrapper


def process_train_labels(x, y):

    train = pd.DataFrame(x)
    if y is None:
        return train, y
    label = pd.DataFrame(y)
    if label.shape[1] == train.shape[0]:
        label = label.T

    assert label.shape[0] == train.shape[0], "Train and label size mismatch"
    return train, label


def _clean_up_temp_folder(path, signal=None, frame=None):
    if signal is not None:
        traceback.print_stack(frame)
    try:
        [
            os_remove(x) if os.path.isfile(x) else shutil.rmtree(x)
            for x in glob.glob(path + "*")
        ]
        os.rmdir(os.path.dirname(path))
    except OSError:
        pass

    # print(frame.print_stack())
    # aulthandler.dump_traceback()
    # print(traceback)


def getValidKwargs(func, argsDict):

    args = set(inspect.getfullargspec(func).args)
    kwargs = set(inspect.getfullargspec(func).kwonlyargs)
    validKwargs = args.union(kwargs)

    validKwargs.discard('self')
    return dict((key, value) for key, value in argsDict.items()
                if key in validKwargs)


def callwithValidKwargs(func, argsDict):
    return func(**getValidKwargs(func, argsDict))


@run_once_per_process
def add_signal(path):
    atexit.register(_clean_up_temp_folder, path)
    # signal.signal(signal.SIGINT, partial(_clean_up_temp_folder, path))
    # signal.signal(signal.SIGTERM, partial(_clean_up_temp_folder, path))
    # faulthandler.register(signal.SIGINT)


def RANDOM_NAME(clean=True, path = "zed_temp"):
    if not os.path.isdir(path):
        os.mkdir(path)
    path = os.path.join(path, "clean_")
    if clean:
        # path = path + str(os.getpgrp()) + "_"
        add_signal(path)
    random_name = str(uuid.uuid4())
    full = path + random_name
    while os.path.isfile(full):
        # print("name_double")
        random_name = str(uuid.uuid4())
        full = path + random_name

    return full



import ctypes
import os
import csv
import shutil
import numpy as np
import pandas as pd
from zedsuite.utils import RANDOM_NAME, process_train_labels, CythonWrapper
cimport cython
from zedsuite.quantizer cimport _quantizer
from libc.stdlib cimport malloc, free
from libc.string cimport strcpy


class Quantizer(CythonWrapper):
    def __init__(
        self,
        helpmsg: bool = False,
        version: bool = False,
        percentage: ctypes.c_double = None,
        pruned: bool = None,
        detrend: ctypes.c_uint = None,
        normalized: bool = None,
        epsilon: ctypes.c_double = -1,
        min_alphabet_size: ctypes.c_uint = 2,
        max_alphabet_size: ctypes.c_uint = 3,
        verbose: ctypes.c_uint = 0,
        output: str = None,
        save_quantized: ctypes.c_uint = None,
        consider_single_symbol: ctypes.c_uint = None,
        n_quantizations: ctypes.c_uint = 1,
        return_failed: bool = True,
        clean: bool = True,
    ) -> None:

        super().__init__()

        self.argc = 1
        self.argv = ['quantizer']

        # Program information:
        if helpmsg:
            self._add_flag("-h")
        if version:
            self._add_flag("-V")

        # Usage of Binary:
        if percentage is not None:
            self._add_param("-x", str(percentage))
        if pruned is not None:
            self._add_param("-r", str(int(pruned)))
        if detrend is not None:
            self._add_param("-d", str(detrend))
        if normalized is not None:
            self._add_param("-n", str(int(normalized)))
        if epsilon is not None:
            self._add_param("-e", str(epsilon))
        if min_alphabet_size is not None:
            self._add_param("-a", str(min_alphabet_size))
        if max_alphabet_size is not None:
            self._add_param("-A", str(max_alphabet_size))
        if verbose is not None:
            self._add_param("-v", str(verbose))
        if output is not None:
            self._add_param("-o", output)
        if save_quantized is not None:
            self._add_param("-q", str(save_quantized))
        if consider_single_symbol is not None:
            self._add_param("-f", str(consider_single_symbol))
        if n_quantizations is not None:
            self._add_param("-M", str(n_quantizations))

        # Additional options
        self._num_quantizations = n_quantizations
        self._return_failed = return_failed
        self._verbose = verbose
        self._fitted = False
        self._clean = clean
        self._feature_order = []
        self._temp_data_dir = RANDOM_NAME(clean=self._clean)

        self.parameters = {}
        self.lib_files = []
        self.training_X = None
        self.supervised = True

        # Results
        self.rc = None

    def run(self) -> bool:
        # convert self.argv to a char** to be passed to main()
        c_argv = <char **> malloc(self.argc * cython.sizeof(ctypes.c_char_p))
        for i, arg in enumerate(self.argv):
            c_argv[i] = <char *> malloc(len(arg) * cython.sizeof(char))
            strcpy(c_argv[i], bytes(arg, 'utf-8'))

        # _quantizer() defined in _quantizer.cc
        if self._verbose > 0:
            print("./", end="")
            [print(arg, end=" ") for arg in self.argv]
            print("\nQUANTIZING")
        self.rc = _quantizer(self.argc, c_argv)
        if self._verbose > 0:
            print("\nQUANTIZED")

        for i in range(self.argc):
            free(c_argv[i])
        free(c_argv)

        return self.rc

    def fit(self, data_dir, label=None, *, force_refit=False):
        if self._num_quantizations == 0:
            return data_dir
        if self._fitted and not force_refit:
            return None
        if (label is None) and (type(data_dir) is not str):
            self.supervised = False
            label = pd.DataFrame([1] * data_dir.shape[0])
        if type(data_dir) is not str:
            if os.path.exists(self._temp_data_dir):
                shutil.rmtree(self._temp_data_dir)
            os.mkdir(self._temp_data_dir)
            data_dir = data_dir.reset_index(drop=True)
            data_dir, label = process_train_labels(data_dir, label)

            # binary expects different dir format for unsupervised/single class case
            classes = set(label[list(label)[0]])
            if len(classes) == 1:
                self.supervised = False
                outfile = "dataset"
                data_dir.reindex(index=label[label[list(label)[0]] == 1].index).to_csv(
                    os.path.join(self._temp_data_dir, outfile),
                    sep=" ",
                    index=False,
                    header=False,
                )
            else:
                file = open(os.path.join(self._temp_data_dir, "library_list"), "w")
                for i in classes:
                    outfile = "train_class_" + str(i)
                    df_i = data_dir.reindex(index=label[label[list(label)[0]] == i].index)
                    df_i.to_csv(
                        os.path.join(self._temp_data_dir, outfile),
                        sep=" ",
                        index=False,
                        header=False,
                    )
                    file.write(
                        outfile + " " + str(i) + " " + str(df_i.shape[0]) + " " + "\n"
                    )
                file.close()
            self.data_dir = self._temp_data_dir

        else:
            self.data_dir = data_dir

        # size = data_dir.shape[0] * data_dir.shape[1]
        # self._sample_size = min(1, 100000 / size)
        self._add_param("-D", self.data_dir)
        self._add_param("-t", "2" if self.supervised else "0")
        self.run()

        # if self._problem_type == 'supervised':
        self._note_lib_files(self.data_dir)

        prune_range_list = []
        detrending_list = []
        normalization_list = []
        partition_list = []

        valid_params_path = os.path.join(self.data_dir, "valid_parameter")
        with open(valid_params_path) as f:
            valid_parameters = f.read().splitlines()
        if self._clean:
            try:
                shutil.rmtree(self._temp_data_dir)
            except:
                pass
        valid_parameters.sort()  # sorts normally by alphabetical order
        valid_parameters.sort(key=len, reverse=True)  # sorts by descending length

        # combined previous two for-loops into one
        parameters = {}
        for param in valid_parameters:
            pr, d, n, pa = self._read_quantizer_params(param)
            key = self._write_quantizer_params(pr, d, n, pa)
            param_set = {
                "prune_range": pr,
                "detrending": d,
                "normalization": n,
                "partition": pa,
            }
            parameters[key] = param_set

            # don't include duplicate quantizations
            if key not in self._feature_order:
                self._feature_order.append(key)

        assert (
            len(self._feature_order) != 0
        ), "Quantization failed, try manual quantization or normalize the data."

        self.parameters = parameters
        self._fitted = True
        return self

    def transform(self, data, *, output_type="matrix"):
        if self._num_quantizations == 0:
            return data
        assert self._fitted, (
            "'fit()' or 'fit_transform()' must be called prior to running 'transform()'"
        )

        if not isinstance(data, pd.DataFrame):
            if not os.path.isdir(self.data_dir):
                os.mkdir(self.data_dir)

        success = []
        for _, name in enumerate(self._feature_order):
            p_dict = self.parameters[name]
            result = self._try_apply_quantizer(
                data, **p_dict
            )
            if result is None:
                success.append(result)
            else:
                success.append(True)
            yield result

        if isinstance(data, pd.DataFrame):
            assert not all(
                v is None for v in success
            ), "All quantizations failed on train"

    def fit_transform(self, data_dir, label=None, *, output_type="matrix", force_refit=False):
        self.fit(data_dir, label=label, force_refit=force_refit)
        if type(data_dir) is not str:
            return self.transform(data_dir)
        X = []
        for lib_file in self.lib_files:
            lib_path = os.path.join(data_dir, lib_file)
            X_ = self.transform(lib_path, output_type=output_type)
            if X_ is None:
                return None
            X.append(X_)
        X_ = np.vstack(X)
        self.training_X = X_
        if output_type == "filename":
            X_ = X
        return X_

    def _note_lib_files(self, data_dir):
        library_list = os.path.join(data_dir, "library_list")
        if os.path.isfile(library_list):
            with open(library_list) as f:
                train_data = [row.split(" ")[:2] for row in f.read().splitlines()]
        else:
            train_data = [["dataset", -1]]
        for lib, _ in train_data:
            self.lib_files.append(lib)

    @staticmethod
    def _detrend(df, *, detrend_level):
        return df.diff(axis=1).dropna(how="all", axis=1)

    @staticmethod
    def _normalize(df):
        df_stdev = df.std(axis=1)
        pos_stdev_0 = df_stdev==0
        df_stdev[pos_stdev_0] = 1

        standard_normal_rows = df.subtract(df.mean(axis=1),
                                           axis=0).divide(df_stdev,
                                                          axis=0)
        return standard_normal_rows

    @staticmethod
    def _prune_func(df, lower_bound, upper_bound):
        for index in df.index:
            X = []
            for val in df.loc[index].values:
                if val <= float(lower_bound) or val >= float(upper_bound):
                    X = np.append(X, val)
            pruned_ = np.empty([1, len(df.loc[index].values) - len(X)])
            pruned_[:] = np.nan
            X = np.append(X, pruned_)
            df.loc[index] = X
        return df

    def _try_apply_quantizer(
        self,
        data,
        *,
        partition,
        prune_range=None,
        detrending=None,
        normalization=None,
        outfile=None,
        verbose=False
    ):
        max_col_len = 0
        if type(data) is str:
            with open(data, "r") as infile:
                csv_reader = csv.reader(infile, delimiter=" ")
                for row in csv_reader:
                    len_ = len(row)
                    if len_ > max_col_len:
                        max_col_len = len_
            unquantized = pd.read_csv(
                data,
                delimiter=" ",
                dtype="float",
                header=None,
                names=range(max_col_len),
            )
        else:
            unquantized = data.copy().astype(float)

        # columns_list = list(unquantized.columns)
        # index_list = unquantized.index
        if prune_range:
            if verbose:
                print("PRUNING")
            unquantized = self._prune_func(unquantized, prune_range[0], prune_range[1])
        # pd.DataFrame(unquantized).to_csv("debug1")
        if detrending:
            if verbose:
                print("DETRENDING")
            unquantized = self._detrend(unquantized, detrend_level=detrending)
        # pd.DataFrame(unquantized).to_csv("debug2")
        if normalization:
            if verbose:
                print("NORMALIZING")
            unquantized = self._normalize(unquantized)
        # unquantized.to_csv("debug3")
        if outfile is None:
            # _outfile = filename
            pass
        else:
            _outfile = outfile
        # if type(data) is not str:
        partition = [float(i) for i in partition]
        # unquantized.dropna(how="all", axis=1).to_csv("unq.csv", sep=",", header=False, index=False)
        # pd.DataFrame(np.digitize(unquantized.values, bins=partition)).to_csv("test.csv", sep=",", header=False, index=False)
        quantized = pd.DataFrame(
            np.digitize(unquantized.values, bins=partition) #, index=data.index
        )
        # quantized[unquantized.isnull()] = np.nan
        if (
            not self._correct_num_symbols(quantized.values, partition)
            and not self._return_failed
        ):
            print("Dropping invalid quantization.")
            return None
        return quantized

        # if not os.path.isfile(_outfile):
        #     print("Failed to apply quantization! Retrying...")
        #     self._try_apply_quantizer(
        #         _outfile,
        #         partition=partition,
        #         prune_range=prune_range,
        #         detrending=detrending,
        #         normalization=normalization,
        #         outfile=outfile,
        #         verbose=verbose,
        #     )
        # return True

    @staticmethod
    def _correct_num_symbols(quantized_matrix, partition):

        expected_num_symbols = len(partition) + 1
        i = 0
        for row in quantized_matrix:
            num_symbols = len(np.unique(row))
            if num_symbols == 1:
                # print("SINGLE SYMBOL STREAM SOMEHOW PASSED CHECK: {}".format(partition))
                pass
            if num_symbols != expected_num_symbols:
                # np.savetxt("foo.csv", quantized_matrix, delimiter=",")
                return False
            i += 1

        return True

    @staticmethod
    def _read_quantizer_params(parameters):

        parameters_ = parameters.split("L")[0]
        prune_range = []
        for index, char in enumerate(parameters_):
            if char == "R":
                for char_ in parameters_[index + 2 :]:
                    if char_ != "]":
                        prune_range.append(char_)
                    else:
                        break
            elif char == "D":
                detrending = int(parameters_[index + 1])
            elif char == "N":
                normalization = int(parameters_[index + 1])
        if prune_range:
            prune_range = "".join(prune_range).split(" ")

        partition = parameters_.split("[")[-1].strip("]").split()
        no_negative_zero_partition = []
        for p in partition:
            if repr(float(p)) == "-0.0":
                p_ = p[1:]
            else:
                p_ = p
            no_negative_zero_partition.append(p_)

        return prune_range, detrending, normalization, no_negative_zero_partition

    @staticmethod
    def _write_quantizer_params(prune_range, detrending, normalization, partition):
        """

        """
        params = []
        if prune_range:
            params.append("R")
            params += prune_range
        if detrending:
            params.append("D")
            params.append(detrending)
        if normalization:
            params.append("N")
            params.append(normalization)
        if partition:
            params.append("P")
            params += str([float(p) for p in partition]).replace(" ", "")
        params_string = "".join([str(p) for p in params])
        return params_string



# details here: https://stackoverflow.com/a/52729181/8728279
import sys
import importlib

# Chooses the right init function
class CythonPackageMetaPathFinder(importlib.abc.MetaPathFinder):
    def __init__(self, name_filter):
        super(CythonPackageMetaPathFinder, self).__init__()
        self.name_filter =  name_filter

    def find_module(self, fullname, path):
        if fullname.startswith(self.name_filter):
            # use this extension-file but PyInit-function of another module:
            return importlib.machinery.ExtensionFileLoader(fullname,__file__)


# injecting custom finder/loaders into sys.meta_path:
def bootstrap_cython_submodules():
    sys.meta_path.append(CythonPackageMetaPathFinder('zedsuite.'))



from os import path
from setuptools import setup, find_packages
from distutils.extension import Extension
from Cython.Build import cythonize

here = path.abspath(path.dirname(__file__))
package_name = "zedsuite"

version = {}
with open("version.py") as fp:
    exec(fp.read(), version)

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


extensions = [Extension(
    name=f"{package_name}.bootstrap",
    sources=[
        f"{package_name}/bootstrap.pyx",
        f"{package_name}/genesess.pyx",
        f"{package_name}/quantizer.pyx",
        f"{package_name}/zutil.pyx",
    ],
    extra_compile_args=[
        "-std=c++11",
        "-O3",
        "-static",
        "-fopenmp",
        "-I./zbase",
        "-I./boost",
        "-Wl,--as-needed",
        "-DCYTHON_PEP489_MULTI_PHASE_INIT=0", # switch multi-phase initialization off which is default in Cython 0.29 for Python>=3.5
    ],
    libraries=[
        "semcrct", "configfile", # recompiled from zbase/lib with -fPIC
        "gsl", "gslcblas",
        "gomp",
        "m",
        "boost_system",
        "boost_thread",
        "boost_program_options",
        "boost_timer",
        "boost_chrono",
    ],
    extra_link_args=[
        "-std=c++11",
        "-fopenmp",
        "-static-libgcc",
        "-static-libstdc++",
        "-O3",
        "-Wl,--as-needed",
    ],
    library_dirs=[
        ".",
        "./zedsuite",
        "./boost/lib",
        "/usr/lib64", # use gsl libs here instead of in zbase
        "./zbase/lib",
    ],
    language="c++"
)]

setup(
    name=package_name,
    author='zed.uchicago.edu',
    author_email='ishanu@uchicago.edu',
    version = str(version['__version__']),
    packages=find_packages(),
    scripts=[],
    url="https://github.com/zeroknowledgediscovery/python_implementations_",
    license='LICENSE',
    description="Cythonized Libraires of ZeD Code",
    keywords=[
        'cython',
        'machine learning',
    ],
    download_url='https://github.com/zeroknowledgediscovery/python_implementations_/archive/'+str(version['__version__'])+'.tar.gz',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    ext_modules=cythonize(
        extensions,
        language_level="3" # set python3
    ),
    install_requires=[
        "Cython>=0.29.23"
        "numpy",
        "pandas",
    ],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6"],
    include_package_data=True,
)
